# simple test using raw arrays instead of numpy
import plumed
import os
import array
import unittest

class Test(unittest.TestCase):
  def test(self):

    p = plumed.Plumed()
    p.cmd("setNatoms",2)
    p.cmd("setLogFile","test.log")
    p.cmd("init")
    p.cmd("readInputLine","d: DISTANCE ATOMS=1,2")
    p.cmd("readInputLine","RESTRAINT ARG=d AT=0 KAPPA=1")

    box=array.array('d',[10,0,0,0,10,0,0,0,10])
    virial=array.array('d',[0,0,0,0,0,0,0,0,0])
    masses=array.array('d',[1,1])
    charges=array.array('d',[0,0])
    forces=array.array('d',[0,0,0,0,0,0])
    positions=array.array('d',[0,0,0,1,2,3])

    p.cmd("setStep",0)
    p.cmd("setBox",box )
    p.cmd("setMasses", masses )
    p.cmd("setCharges", charges )
    p.cmd("setPositions", positions )
    p.cmd("setForces", forces )
    p.cmd("setVirial", virial )
    p.cmd("calc")

    bias=array.array('d',[0])

    p.cmd("getBias", bias )

    self.assertAlmostEqual( bias[0]  ,   7.0)
    self.assertAlmostEqual( forces[0],   1.0)
    self.assertAlmostEqual( forces[1],   2.0)
    self.assertAlmostEqual( forces[2],   3.0)
    self.assertAlmostEqual( forces[3], - 1.0)
    self.assertAlmostEqual( forces[4], - 2.0)
    self.assertAlmostEqual( forces[5], - 3.0)

if __name__ == "__main__":
    unittest.main()




import plumedCommunications as PLMD

# import plumedUtilities
log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

plumedInit = {
    "COMPONENTS": {
        "mass0": PLMD.defaults.COMPONENT_NODEV,
        "mass1": PLMD.defaults.COMPONENT_NODEV,
        "charge0": PLMD.defaults.COMPONENT_NODEV,
        "charge1": PLMD.defaults.COMPONENT_NODEV,
    }
}


def mypytest(action: PLMD.PythonCVInterface):
    ret = {
        "mass0": action.mass(0),
        "charge0": action.charge(0),
        "mass1": action.mass(1),
        "charge1": action.charge(1),
    }

    return ret



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)


def init(action: PLMD.PythonCVInterface):
    return {"COMPONENTS": {"val":{"period": ["0", 1.3]}}}

initForF={"Value": PLMD.defaults.COMPONENT_NODEV}


def mypytest(action: PLMD.PythonCVInterface):
    ret = [action.getStep()]

    return ret

def function(action: PLMD.PythonFunction):
    arg = [action.argument(0)]
    return arg[0]*arg[0]


import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

initForF = {
    "COMPONENTS": {
        "x": PLMD.defaults.COMPONENT_NODEV,
        "y": PLMD.defaults.COMPONENT_NODEV,
        "z": PLMD.defaults.COMPONENT_NODEV,
    }
}


def function(action: PLMD.PythonFunction):
    arg = action.arguments()

    return {
        "x": arg[0],
        "y": arg[1],
        "z": arg[2],
    }



# checked and formatted with ruff 0.6.9


def getPythonCVInterface():
    """returns the location of the pycv shared object"""
    # older version
    # import inspect
    # import pycv

    # path_of_this = inspect.getfile(pycv)
    import importlib.util

    path_of_this = importlib.util.find_spec("pycv").origin
    path_of_lib = path_of_this[: path_of_this.rfind("/")]
    return path_of_lib + "/PythonCVInterface.so"


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser(
        prog="pyCV",
        description="""shows the path for the pycv shared object.

        Just run this with no arguments to see the path.
        """,
    )
    parser.add_argument(
        "-p",
        "--plumed",
        help="print on screen an example plumed.dat",
        action="store_true",
    )
    args = parser.parse_args()

    if args.plumed:
        print(
            f"""LOAD FILE={getPythonCVInterface()}
cvPy:PYCVINTERFACE IMPORT=mypycv
fPy: PYFUNCTION IMPORT=mypycvfunc ARG=cvPy
PRINT FILE=colvar.out ARG=*"""
        )
        return 0

    print(getPythonCVInterface())
    return 0



import unittest
import numpy as np

from utilities_for_test import *

import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

# CALCULATE is tested "by default"


class TestFunctionCalls(unittest.TestCase):
    def test_INIT(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("traj.xyz")
            )
            plmd = preparePlumed(num_atoms)
            d1 = create_plumed_var(plmd, "d1", "DISTANCE ATOMS=1,2")
            d2 = create_plumed_var(plmd, "d2", "DISTANCE ATOMS=1,3")

            fPy = create_plumed_var(
                plmd,
                "fPY",
                "PYFUNCTION IMPORT=pyfuncINITfunc INIT=initForF CALCULATE=function ARG=d1,d2",
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            for step in range(num_frames):
                plmd.cmd("setStep", step)
                plmd.cmd("setBox", box)
                plmd.cmd("setMasses", masses)
                plmd.cmd("setCharges", charges)
                plmd.cmd("setPositions", traj[step])
                plmd.cmd("setForces", forces)
                plmd.cmd("setVirial", virial)
                plmd.cmd("calc")

                np.testing.assert_almost_equal(fPy, d1 * d2, decimal=4)

    def test_INITDICT(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("traj.xyz")
            )
            plmd = preparePlumed(num_atoms)

            plmd.cmd("readInputLine", "dc: DISTANCE ATOMS=1,2 COMPONENTS")
            shape = np.zeros(1, dtype=np.int_)
            plmd.cmd("getDataRank " + "dc.x", shape)
            plmd.cmd("getDataRank " + "dc.y", shape)
            plmd.cmd("getDataRank " + "dc.z", shape)
            dcx = np.zeros((1))
            dcy = np.zeros((1))
            dcz = np.zeros((1))
            plmd.cmd("setMemoryForData dc.x", dcx)
            plmd.cmd("setMemoryForData dc.y", dcy)
            plmd.cmd("setMemoryForData dc.z", dcz)

            d = create_plumed_var(plmd, "d", "DISTANCE ATOMS=1,2")

            fPy = create_plumed_var(
                plmd,
                "fPY",
                "PYFUNCTION IMPORT=pyfuncINITdict INIT=initForF CALCULATE=function ARG=dc.x,dc.y,dc.z,d",
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            for step in range(num_frames):
                plmd.cmd("setStep", step)
                plmd.cmd("setBox", box)
                plmd.cmd("setMasses", masses)
                plmd.cmd("setCharges", charges)
                plmd.cmd("setPositions", traj[step])
                plmd.cmd("setForces", forces)
                plmd.cmd("setVirial", virial)
                plmd.cmd("calc")

                np.testing.assert_almost_equal(
                    fPy, np.abs(np.sqrt(dcx**2 + dcy**2 + dcz**2) - d[0]), decimal=4
                )


if __name__ == "__main__":
    # Output to four decimal places only
    np.set_printoptions(precision=4)
    unittest.main()



import unittest
import plumed
import filecmp
import os
from contextlib import contextmanager

@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)

class Test(unittest.TestCase):
  def test1(self):
    with cd('test/'):
        d=plumed.read_as_pandas("COLVAR",enable_constants='columns')
        with open("dataframe","wt") as f:
            print(d,file=f)
        self.assertTrue(filecmp.cmp("dataframe","dataframe.ref"))

  def test2(self):
    with cd('test/'):
        d=plumed.read_as_pandas("COLVAR_gzipped.gz",enable_constants='columns')
        with open("dataframe","wt") as f:
            print(d,file=f)
        self.assertTrue(filecmp.cmp("dataframe","dataframe.ref"))

  def test3(self):
    with cd('test/'):
        i=0
        for d in plumed.read_as_pandas("COLVAR",chunksize=4,enable_constants='columns'):
            with open("dataframe."+str(i),"wt") as f:
                print(d,file=f)
            self.assertTrue(filecmp.cmp("dataframe."+str(i),"dataframe."+str(i)+".ref"))
            i=i+1

  def test4(self):
    with cd('test/'):
        i=0
        for d in plumed.read_as_pandas("COLVAR_gzipped.gz",chunksize=4,enable_constants='columns'):
            with open("dataframe."+str(i),"wt") as f:
                print(d,file=f)
            self.assertTrue(filecmp.cmp("dataframe."+str(i),"dataframe."+str(i)+".ref"))
            i=i+1

  def test5(self):
    with cd('test/'):
        d=plumed.read_as_pandas("COLVAR",enable_constants='metadata')
        with open("dataframe_noconst","wt") as f:
            print(d,file=f)
        self.assertTrue(filecmp.cmp("dataframe_noconst","dataframe_noconst.ref"))
        self.assertEqual(d.plumed_constants[0][0],"a")
        self.assertEqual(d.plumed_constants[0][2],"pi")

  def test6(self):
    with cd('test/'):
        i=0
        for d in plumed.read_as_pandas("COLVAR",chunksize=4,enable_constants='metadata'):
            with open("dataframe_noconst."+str(i),"wt") as f:
                print(d,file=f)
            self.assertTrue(filecmp.cmp("dataframe_noconst."+str(i),"dataframe_noconst."+str(i)+".ref"))
            self.assertEqual(d.plumed_constants[0][0],"a")
            self.assertEqual(d.plumed_constants[0][2],"pi")
            i=i+1

  def test7(self):
    with cd('test/'):
        d=plumed.read_as_pandas("COLVAR")
        plumed.write_pandas(d,"COLVAR_write1")
        self.assertTrue(filecmp.cmp("COLVAR_write1","COLVAR_write.ref"))
        d=plumed.read_as_pandas("COLVAR",index_col='time')
        plumed.write_pandas(d,"COLVAR_write2")
        self.assertTrue(filecmp.cmp("COLVAR_write2","COLVAR_write.ref"))
        d=plumed.read_as_pandas("COLVAR",index_col=('time','psi'))
        try:
            plumed.write_pandas(d,"COLVAR_write3")
            self.assertTrue(False)
        except TypeError:
            pass

if __name__ == "__main__":
    unittest.main()



#/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   Copyright (c) 2011-2016 The plumed team
#   (see the PEOPLE file at the root of the distribution for a list of names)
#
#   See http://www.plumed.org for more information.
#
#   This file is part of plumed, version 2.
#
#   plumed is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   plumed is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with plumed.  If not, see <http://www.gnu.org/licenses/>.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#
# This create cython wrappers to the main bits of the PLUMED libraray
#

cdef extern from "Plumed.h":
    ctypedef struct plumed:
        pass # ignore content
    ctypedef struct plumed_safeptr:
        const void* ptr
        size_t nelem
        size_t* shape
        size_t flags
        # ignore other
    ctypedef struct plumed_nothrow_handler:
        void* ptr
        void (*handler)(void*,int,const char*,const void*)
    ctypedef struct plumed_error_filesystem_path:
        size_t numbytes
        void* ptr
    ctypedef struct plumed_error:
        int code
        const char* what
        void* nested
        int error_code
        plumed_error_filesystem_path path1
        plumed_error_filesystem_path path2
        # ignore other members
    # C functions declared as nogil
    # see https://github.com/plumed/plumed2/pull/1129#issuecomment-2410867829
    void plumed_cmd_safe_nothrow(plumed p,const char*key,plumed_safeptr safe,plumed_nothrow_handler nothrow) nogil
    void plumed_error_set(void*ptr,int code,const char*what,const void* opt) nogil
    void plumed_error_init(plumed_error* error) nogil
    void plumed_error_finalize(plumed_error error) nogil
    plumed plumed_create() nogil
    plumed plumed_create_dlopen(const char*path) nogil
    plumed plumed_create_invalid() nogil
    void plumed_finalize(plumed p) nogil
    int plumed_valid(plumed p) nogil



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

initForF={"Value": PLMD.defaults.COMPONENT_NODEV}

def function(action: PLMD.PythonFunction):
    arg = [action.argument(0),
           action.argument(1)
           ]
    return arg[0]*arg[0]*arg[1]


import pytest

import hillclimber as hc


def test_opes_metad_basic_adaptive_sigma(small_ethanol_water):
    """Test basic OPES_METAD with adaptive sigma."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        explore_mode=False,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_fixed_sigma(small_ethanol_water):
    """Test OPES_METAD with fixed sigma value."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma=0.2)

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=0.2 FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_explore_mode(small_ethanol_water):
    """Test OPES_METAD_EXPLORE variant."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        explore_mode=True,  # Enable exploration mode
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD_EXPLORE ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_two_cvs_same_sigma(small_ethanol_water):
    """Test OPES_METAD with two CVs using same sigma."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    torsion_cv = hc.TorsionCV(
        atoms=hc.IndexSelector(indices=[[0, 1, 2, 3]]),
        prefix="phi",
    )

    opes_bias1 = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")
    opes_bias2 = hc.OPESBias(cv=torsion_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias1, opes_bias2],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "phi: TORSION ATOMS=1,2,3,4",
        "opes: OPES_METAD ARG=d12,phi PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_two_cvs_different_sigma(small_ethanol_water):
    """Test OPES_METAD with two CVs using different sigma values."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    torsion_cv = hc.TorsionCV(
        atoms=hc.IndexSelector(indices=[[0, 1, 2, 3]]),
        prefix="phi",
    )

    opes_bias1 = hc.OPESBias(cv=distance_cv, sigma=0.1)
    opes_bias2 = hc.OPESBias(cv=torsion_cv, sigma=0.2)

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias1, opes_bias2],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "phi: TORSION ATOMS=1,2,3,4",
        "opes: OPES_METAD ARG=d12,phi PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=0.1,0.2 FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_with_optional_parameters(small_ethanol_water):
    """Test OPES_METAD with optional parameters."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        biasfactor=10.0,
        adaptive_sigma_stride=5000,
        sigma_min=0.01,
        compression_threshold=0.5,
        file="MY_KERNELS",
        calc_work=True,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=MY_KERNELS COMPRESSION_THRESHOLD=0.5 BIASFACTOR=10.0 ADAPTIVE_SIGMA_STRIDE=5000 SIGMA_MIN=0.01 CALC_WORK",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_with_walkers_mpi(small_ethanol_water):
    """Test OPES_METAD with multiple walkers."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        walkers_mpi=True,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0 WALKERS_MPI",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_with_state_files(small_ethanol_water):
    """Test OPES_METAD with state file restart."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        state_wfile="STATE",
        state_wstride=1000,
        state_rfile="OLD_STATE",
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0 STATE_WFILE=STATE STATE_RFILE=OLD_STATE STATE_WSTRIDE=1000",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_with_actions(small_ethanol_water):
    """Test OPES_METAD with additional actions (restraints, walls, print)."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    # Create restraint and wall
    restraint = hc.RestraintBias(cv=distance_cv, kappa=100.0, at=2.5)
    upper_wall = hc.UpperWallBias(cv=distance_cv, at=5.0, kappa=50.0)

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[restraint, upper_wall],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "d12_restraint: RESTRAINT ARG=d12 KAPPA=100.0 AT=2.5",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "d12_uwall: UPPER_WALLS ARG=d12 AT=5.0 KAPPA=50.0 EXP=2",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_metad_with_flush(small_ethanol_water):
    """Test OPES_METAD with FLUSH parameter."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
        flush=100,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "opes: OPES_METAD ARG=d12 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=ADAPTIVE FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
        "FLUSH STRIDE=100",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected


def test_opes_duplicate_cv_prefix(small_ethanol_water):
    """Test that duplicate CV prefixes are detected."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector[0], "com"),
        x2=hc.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma="ADAPTIVE")

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias, opes_bias],  # duplicate entry
        actions=[],
        model=None,  # type: ignore
    )

    with pytest.raises(ValueError, match="Duplicate CV prefix found: d12"):
        opes_model.to_plumed(small_ethanol_water[0])


def test_opes_metad_diagonal_strategy(small_ethanol_water):
    """Test OPES_METAD with diagonal/corresponding strategy - pair by index."""
    x1_selector = hc.SMILESSelector(smiles="CCO")
    x2_selector = hc.SMILESSelector(smiles="O")

    distance_cv = hc.DistanceCV(
        x1=hc.VirtualAtom(x1_selector, "com"),
        x2=hc.VirtualAtom(x2_selector, "com"),
        prefix="d12",
        pairwise="diagonal",
    )

    opes_bias = hc.OPESBias(cv=distance_cv, sigma=0.15)

    opes_config = hc.OPESConfig(
        barrier=40.0,
        pace=500,
        temp=300.0,
    )

    opes_model = hc.OPESModel(
        config=opes_config,
        data=small_ethanol_water,
        bias_cvs=[opes_bias],
        actions=[],
        model=None,  # type: ignore
    )

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1_0: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x1_1: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d12_x2_0: COM ATOMS=19,20,21",
        "d12_x2_1: COM ATOMS=22,23,24",
        "d12_0: DISTANCE ATOMS=d12_x1_0,d12_x2_0",
        "d12_1: DISTANCE ATOMS=d12_x1_1,d12_x2_1",
        "opes: OPES_METAD ARG=d12_0,d12_1 PACE=500 BARRIER=40.0 TEMP=300.0 SIGMA=0.15 FILE=KERNELS COMPRESSION_THRESHOLD=1.0",
    ]

    assert opes_model.to_plumed(small_ethanol_water) == expected



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

init = dict(
    COMPONENTS=dict(
        nonPeriodic=dict(period=None),
        Periodic={"period": ["0", "1.3"]},
        PeriodicPI={"period": ["0", "pi"]},
    ),
    ATOMS="1,2",
)


def mypytest(action: PLMD.PythonCVInterface):
    ret = {
        "nonPeriodic": action.getStep(),
        "Periodic": action.getStep(),
        "PeriodicPI": action.getStep(),
    }

    return ret



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import numpy as np
import plumedCommunications

# import plumedUtilities
log = open("pydist.log", "w")

print("Imported my pydist.", file=log)

plumedInit={"Value":plumedCommunications.defaults.COMPONENT_NODEV}

def pydist(action: plumedCommunications.PythonCVInterface):
    at: np.ndarray = action.getPositions()

    # print(at0,file=log)
    d = at[0] - at[1]
    d = action.getPbc().apply([d])
    assert d.shape[1] == 3, "d is not a (*,3) array"
    d = np.linalg.norm(d[0])

    # print(f"{at1},{at2}",file=log)
    return d



#/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   Copyright (c) 2011-2016 The plumed team
#   (see the PEOPLE file at the root of the distribution for a list of names)
#
#   See http://www.plumed.org for more information.
#
#   This file is part of plumed, version 2.
#
#   plumed is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   plumed is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with plumed.  If not, see <http://www.gnu.org/licenses/>.
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#
# This python routine builds an interface between plumed and python using cython
#
from setuptools import setup
from setuptools import Extension
import subprocess
import os
import os.path
import sys
from shutil import copyfile
import platform
from distutils.sysconfig import get_config_var
from distutils.version import LooseVersion

if sys.version_info < (3,):
    raise ImportError("PLUMED 2.6 only supports Python 3")

def is_platform_mac():
    return sys.platform == 'darwin'

if os.getenv("plumed_macports") is not None:
    copyfile("../VERSION.txt","PLUMED_VERSION")
    try:
        os.mkdir("include")
    except OSError:
        pass
    copyfile("../src/wrapper/Plumed.h","include/Plumed.h")

plumedname = os.getenv("plumed_program_name")
if plumedname is None:
    plumedname = "plumed"

plumedversion = os.getenv("plumed_version")
if plumedversion is None:
    plumedversion = subprocess.check_output(['grep','-v','#','./PLUMED_VERSION']).decode("utf-8").rstrip()

print( "Module name " + plumedname )
print( "Version number " + plumedversion )

extra_compile_args=['-D__PLUMED_HAS_DLOPEN','-D__PLUMED_WRAPPER_LINK_RUNTIME=1','-D__PLUMED_WRAPPER_IMPLEMENTATION=1','-D__PLUMED_WRAPPER_EXTERN=0']

defaultkernel=os.getenv("plumed_default_kernel")
if defaultkernel is not None:
    extra_compile_args.append("-D__PLUMED_DEFAULT_KERNEL=" + os.path.abspath(defaultkernel))
    print( "Hardcoded PLUMED_KERNEL " + os.path.abspath(defaultkernel))

plumed_disable_rtld_deepbind=os.getenv("plumed_disable_rtld_deepbind")
if plumed_disable_rtld_deepbind is not None:
    extra_compile_args.append("-D__PLUMED_WRAPPER_ENABLE_RTLD_DEEPBIND=0")
    print( "Disabling RTLD_DEEPBIND")

# Fixes problem with compiling the PYTHON interface in Mac OS 10.14 and higher.
# Sets the deployment target to 10.9 when compiling on version 10.9 and above,
# overriding distutils behaviour to target the Mac OS version python was built for.
# This can be overridden by setting MACOSX_DEPLOYMENT_TARGET before compiling the
# python interface.
# This fix is taken from https://github.com/pandas-dev/pandas/pull/24274/files
if is_platform_mac():
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        current_system = LooseVersion(platform.mac_ver()[0])
        python_target = LooseVersion(get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < '10.9' and current_system >= '10.9':
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'

def readme():
    with open('README.rst') as f:
        return f.read()


try:
    include_dirs=[os.environ["plumed_include_dir"]]
except KeyError:
    include_dirs=["./include"]

# allow one to force using cython with env var plumed_force_cython=yes
USE_CYTHON = False
try:
    if(os.environ["plumed_force_cython"]=="yes"):
        print('plumed_force_cython=yes')
        USE_CYTHON = True
except KeyError:
    pass

# if plumed.c is available, do not need cython
if not USE_CYTHON:
    if not os.path.isfile("plumed.c"):
        print('plumed.c not found, cython is needed')
        USE_CYTHON = True

# try to import cython
if USE_CYTHON:
    try:
        print('importing cython')
        from Cython.Build import cythonize
        extension="pyx"
    except ImportError:
        print('Error: building ' + plumedname + ' requires cython. Please install it first with pip install cython')
        sys.exit(-1)
else:
    print('using available plumed.c file')
    extension="c"

ext_modules=[Extension(
     name=plumedname,
     sources=["plumed." + extension],
     language="c",
     include_dirs=include_dirs,
     extra_compile_args=extra_compile_args
  )]

if USE_CYTHON:
    ext_modules=cythonize(ext_modules,language_level=3)

setup(
  name=plumedname,
  version=plumedversion,
  description='Python interface to PLUMED',
  long_description=readme(),
  classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics',
          'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
  ],
  author='Gareth A. Tribello',
  author_email='plumed-users@googlegroups.com',
  url='http://www.plumed.org',
  ext_modules = ext_modules,
  zip_safe=False,
  test_suite='nose.collector',
  python_requires='>=3'
)



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import numpy as np
import plumedCommunications

# import plumedUtilities
log = open("pydist.log", "w")

print("Imported my pydist.", file=log)
print(plumedCommunications.defaults.COMPONENT, file=log)

plumedInit = dict(
    COMPONENTS=dict(
        aix=plumedCommunications.defaults.COMPONENT,
        aiy=plumedCommunications.defaults.COMPONENT,
        aiz=plumedCommunications.defaults.COMPONENT,
        bix=plumedCommunications.defaults.COMPONENT,
        biy=plumedCommunications.defaults.COMPONENT,
        biz=plumedCommunications.defaults.COMPONENT,
        cix=plumedCommunications.defaults.COMPONENT,
        ciy=plumedCommunications.defaults.COMPONENT,
        ciz=plumedCommunications.defaults.COMPONENT,
    )
)


def pyInvBox(action: plumedCommunications.PythonCVInterface):    
    invBox = action.getPbc().getInvBox()
    print(f"{invBox=}", file=log)
    ret = {}
    for i, name in enumerate(["a", "b", "c"]):
        for j, coord in enumerate(["x", "y", "z"]):
            ret[f"{name}i{coord}"] = (invBox[i, j], np.zeros((1, 3)))
    print(ret, file=log)
    return ret



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)


def init(action: PLMD.PythonCVInterface):
    return {"COMPONENTS": {"val":{"period": ["0", 1.3]}}}
    return {"Value": {"period": ["0", "1.3"]}}


def mypytest(action: PLMD.PythonCVInterface):
    ret = [action.getStep()]

    return ret



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import numpy as np
import plumedCommunications

# import plumedUtilities
log = open("pydist.log", "w")

print("Imported my pydist.", file=log)

plumedInit = {
    "COMPONENTS": {
        "ax": plumedCommunications.defaults.COMPONENT,
        "ay": plumedCommunications.defaults.COMPONENT,
        "az": plumedCommunications.defaults.COMPONENT,
        "bx": plumedCommunications.defaults.COMPONENT,
        "by": plumedCommunications.defaults.COMPONENT,
        "bz": plumedCommunications.defaults.COMPONENT,
        "cx": plumedCommunications.defaults.COMPONENT,
        "cy": plumedCommunications.defaults.COMPONENT,
        "cz": plumedCommunications.defaults.COMPONENT,
    }
}


def pyBox(action: plumedCommunications.PythonCVInterface):
    d = action.getPbc().getBox()
    print(f"{d=}", file=log)
    ret = {}
    grad = {}
    for i, name in enumerate(["a", "b", "c"]):
        for j, coord in enumerate(["x", "y", "z"]):
            ret[f"{name}{coord}"] = (d[i, j], np.zeros((1, 3)))
    print(ret, file=log)
    return ret



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

initForF = {
    "COMPONENTS": {
        "difference": PLMD.defaults.COMPONENT_NODEV,
        "bringBackInPbc": PLMD.defaults.COMPONENT_NODEV,
    }
}


def function(action: PLMD.PythonFunction):
    arg = action.arguments()

    return {
        "difference": action.difference(0, action.getStep(), arg[0]),
        "bringBackInPbc": action.bringBackInPbc(0, arg[0] * action.getStep()),
    }



import unittest
import numpy as np

from utilities_for_test import *

import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class TestPyCVCalls(unittest.TestCase):
    def test_INIT(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, _, num_atoms, box, virial, masses, forces, charges = setUpTraj(
                "traj.xyz"
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd, "cvPy", "PYCVINTERFACE ATOMS=4 IMPORT=justInit INIT=myInit"
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            step = 0

            plmd.cmd("setStep", step)
            plmd.cmd("setBox", box)
            plmd.cmd("setMasses", masses)
            plmd.cmd("setCharges", charges)
            plmd.cmd("setPositions", traj[step])
            plmd.cmd("setForces", forces)
            plmd.cmd("setVirial", virial)
            plmd.cmd("calc")
            np.testing.assert_almost_equal(cvPy, 10.0, decimal=4)

    def test_INITDICT(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, _, num_atoms, box, virial, masses, forces, charges = setUpTraj(
                "traj.xyz"
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd, "cvPy", "PYCVINTERFACE ATOMS=4 IMPORT=justInitDict"
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            step = 0

            plmd.cmd("setStep", step)
            plmd.cmd("setBox", box)
            plmd.cmd("setMasses", masses)
            plmd.cmd("setCharges", charges)
            plmd.cmd("setPositions", traj[step])
            plmd.cmd("setForces", forces)
            plmd.cmd("setVirial", virial)
            plmd.cmd("calc")
            np.testing.assert_almost_equal(cvPy, 10.0, decimal=4)

    def test_UPDATE(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, _, num_atoms, box, virial, masses, forces, charges = setUpTraj(
                "traj.xyz"
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd, "cvPy", "PYCVINTERFACE ATOMS=4 IMPORT=justUpdate UPDATE=myUpdate"
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            step = 0

            plmd.cmd("setStep", step)
            plmd.cmd("setBox", box)
            plmd.cmd("setMasses", masses)
            plmd.cmd("setCharges", charges)
            plmd.cmd("setPositions", traj[step])
            plmd.cmd("setForces", forces)
            plmd.cmd("setVirial", virial)
            plmd.cmd("calc")
            np.testing.assert_almost_equal(cvPy, 10.0, decimal=4)

    def test_PREPARE(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, _, num_atoms, box, virial, masses, forces, charges = setUpTraj(
                "traj.xyz"
            )
            plmd = preparePlumed(num_atoms)
            # atoms=4 but the module choses 1
            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=4 IMPORT=justPrepare PREPARE=plumedPrepare",
            )
            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file

            step = 0

            plmd.cmd("setStep", step)
            plmd.cmd("setBox", box)
            plmd.cmd("setMasses", masses)
            plmd.cmd("setCharges", charges)
            plmd.cmd("setPositions", traj[step])
            plmd.cmd("setForces", forces)
            plmd.cmd("setVirial", virial)
            plmd.cmd("calc")

            np.testing.assert_almost_equal(cvPy, 5.0, decimal=4)


if __name__ == "__main__":
    # Output to four decimal places only
    np.set_printoptions(precision=4)
    unittest.main()



import pytest

import hillclimber as pn


def test_distance_cv_corresponding_strategy(small_ethanol_water):
    """Test diagonal/corresponding strategy - pair by index."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector, "com"),
        x2=pn.VirtualAtom(x2_selector, "com"),
        prefix="d12",
        pairwise="diagonal",
    )

    biased_distance_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=10.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=0.5,
        pace=150,
        biasfactor=10.0,
        temp=300.0,
        file="HILLS",
        flush=None,
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance_cv],
        actions=[pn.PrintAction(cvs=[distance_cv], stride=100)],
        model=None,  # type: ignore
    )

    assert meta_d_model.to_plumed(small_ethanol_water) == [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1_0: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x1_1: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d12_x2_0: COM ATOMS=19,20,21",
        "d12_x2_1: COM ATOMS=22,23,24",
        "d12_0: DISTANCE ATOMS=d12_x1_0,d12_x2_0",
        "d12_1: DISTANCE ATOMS=d12_x1_1,d12_x2_1",
        "metad: METAD ARG=d12_0,d12_1 HEIGHT=0.5 PACE=150 TEMP=300.0 FILE=HILLS BIASFACTOR=10.0 SIGMA=0.1,0.1 GRID_MIN=0.0,0.0 GRID_MAX=10.0,10.0 GRID_BIN=100,100",
        "PRINT ARG=d12_0,d12_1 STRIDE=100 FILE=COLVAR",
    ]


def test_distance_cv_first_strategy(small_ethanol_water):
    """Test default FIRST strategy - only first groups."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    biased_distance_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=10.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=0.5,
        pace=150,
        biasfactor=10.0,
        temp=300.0,
        file="HILLS",
        flush=100,
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance_cv],
        actions=[pn.PrintAction(cvs=[distance_cv], stride=100)],
        model=None,  # type: ignore
    )

    assert meta_d_model.to_plumed(small_ethanol_water) == [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
        "metad: METAD ARG=d12 HEIGHT=0.5 PACE=150 TEMP=300.0 FILE=HILLS BIASFACTOR=10.0 SIGMA=0.1 GRID_MIN=0.0 GRID_MAX=10.0 GRID_BIN=100",
        "PRINT ARG=d12 STRIDE=100 FILE=COLVAR",
        "FLUSH STRIDE=100",
    ]


def test_duplicate_cv_prefix(small_ethanol_water):
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d12",
    )

    biased_distance_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=10.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=0.5,
        pace=150,
        biasfactor=10.0,
        temp=300.0,
        file="HILLS",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance_cv, biased_distance_cv],  # duplicate entry
        actions=[],  # PrintCVAction is automatically added
        model=None,  # type: ignore
    )

    with pytest.raises(ValueError, match="Duplicate CV prefix found: d12"):
        meta_d_model.to_plumed(small_ethanol_water)


def test_cv_used_in_bias_and_wall(small_ethanol_water):
    """Test that a CV used in both bias_cvs and actions (wall) is not duplicated."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    # Create a distance CV
    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    # Use the same CV in both bias_cvs and actions
    biased_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.2,
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=300,
    )

    upper_wall = pn.UpperWallBias(
        cv=distance_cv,
        at=4.5,
        kappa=1000.0,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.2,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_cv],
        actions=[upper_wall],
        model=None,  # type: ignore
    )

    result = meta_d_model.to_plumed(small_ethanol_water)

    # Check that the CV is defined only once
    cv_definitions = [line for line in result if line.startswith("d:")]
    assert len(cv_definitions) == 1, (
        f"Expected 1 CV definition, got {len(cv_definitions)}: {cv_definitions}"
    )

    # Check that both METAD and UPPER_WALLS are present
    assert any("metad: METAD" in line for line in result)
    assert any("UPPER_WALLS" in line for line in result)

    # Expected structure
    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2: COM ATOMS=19,20,21",
        "d: DISTANCE ATOMS=d_x1,d_x2",
        "metad: METAD ARG=d HEIGHT=1.2 PACE=500 TEMP=300.0 FILE=HILLS BIASFACTOR=10.0 SIGMA=0.2 GRID_MIN=0.0 GRID_MAX=5.0 GRID_BIN=300",
        "d_uwall: UPPER_WALLS ARG=d AT=4.5 KAPPA=1000.0 EXP=2",
    ]

    assert result == expected


def test_cv_conflict_detection(small_ethanol_water):
    """Test that using the same label with different CV definitions raises an error."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    # Create two different CVs with the same prefix
    distance_cv1 = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    distance_cv2 = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[1], "com"),  # Different group!
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",  # Same prefix!
    )

    biased_cv = pn.MetadBias(
        cv=distance_cv1,
        sigma=0.2,
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=300,
    )

    # Using a different CV with the same prefix in actions should raise an error
    upper_wall = pn.UpperWallBias(
        cv=distance_cv2,
        at=4.5,
        kappa=1000.0,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.2,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_cv],
        actions=[upper_wall],
        model=None,  # type: ignore
    )

    # This should raise an error because d_x1 has different definitions
    with pytest.raises(ValueError, match="Conflicting definitions for label 'd_x1'"):
        meta_d_model.to_plumed(small_ethanol_water)


def test_adaptive_geom(small_ethanol_water):
    """Test that adaptive='GEOM' is correctly written to PLUMED output."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    biased_distance_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.0,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
        adaptive="GEOM",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance_cv],
        model=None,  # type: ignore
    )

    result = meta_d_model.to_plumed(small_ethanol_water)

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2: COM ATOMS=19,20,21",
        "d: DISTANCE ATOMS=d_x1,d_x2",
        "metad: METAD ARG=d HEIGHT=1.0 PACE=500 TEMP=300.0 FILE=HILLS ADAPTIVE=GEOM BIASFACTOR=10.0 SIGMA=0.1 GRID_MIN=0.0 GRID_MAX=5.0 GRID_BIN=100",
    ]

    assert result == expected


def test_adaptive_diff(small_ethanol_water):
    """Test that adaptive='DIFF' is correctly written to PLUMED output."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    biased_distance_cv = pn.MetadBias(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.0,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
        adaptive="DIFF",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance_cv],
        model=None,  # type: ignore
    )

    result = meta_d_model.to_plumed(small_ethanol_water)

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2: COM ATOMS=19,20,21",
        "d: DISTANCE ATOMS=d_x1,d_x2",
        "metad: METAD ARG=d HEIGHT=1.0 PACE=500 TEMP=300.0 FILE=HILLS ADAPTIVE=DIFF BIASFACTOR=10.0 SIGMA=0.1 GRID_MIN=0.0 GRID_MAX=5.0 GRID_BIN=100",
    ]

    assert result == expected


def test_adaptive_multiple_cvs_same_sigma(small_ethanol_water):
    """Test that adaptive with multiple CVs and same sigma writes single sigma value."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    # Create two different CVs with the same sigma
    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    torsion_cv = pn.TorsionCV(
        atoms=pn.IndexSelector(indices=[[0, 1, 3, 4]]),
        prefix="t",
    )

    biased_distance = pn.MetadBias(
        cv=distance_cv,
        sigma=0.2,  # Same sigma
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=100,
    )

    biased_torsion = pn.MetadBias(
        cv=torsion_cv,
        sigma=0.2,  # Same sigma
        grid_min="-pi",
        grid_max="pi",
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.0,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
        adaptive="DIFF",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance, biased_torsion],
        model=None,  # type: ignore
    )

    result = meta_d_model.to_plumed(small_ethanol_water)

    expected = [
        "UNITS LENGTH=A TIME=0.001 ENERGY=96.48533288249877",
        "d_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2: COM ATOMS=19,20,21",
        "d: DISTANCE ATOMS=d_x1,d_x2",
        "t: TORSION ATOMS=1,2,4,5",
        "metad: METAD ARG=d,t HEIGHT=1.0 PACE=500 TEMP=300.0 FILE=HILLS ADAPTIVE=DIFF BIASFACTOR=10.0 SIGMA=0.2 GRID_MIN=0.0,-pi GRID_MAX=5.0,pi GRID_BIN=100,100",
    ]

    assert result == expected


def test_adaptive_multiple_cvs_different_sigma_raises_error(small_ethanol_water):
    """Test that adaptive with multiple CVs and different sigmas raises an error."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    # Create two different CVs with different sigma values
    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),
        x2=pn.VirtualAtom(x2_selector[0], "com"),
        prefix="d",
    )

    torsion_cv = pn.TorsionCV(
        atoms=pn.IndexSelector(indices=[[0, 1, 3, 4]]),
        prefix="t",
    )

    biased_distance = pn.MetadBias(
        cv=distance_cv,
        sigma=0.2,  # Different sigma
        grid_min=0.0,
        grid_max=5.0,
        grid_bin=100,
    )

    biased_torsion = pn.MetadBias(
        cv=torsion_cv,
        sigma=0.3,  # Different sigma!
        grid_min="-pi",
        grid_max="pi",
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=1.0,
        pace=500,
        biasfactor=10.0,
        temp=300.0,
        adaptive="GEOM",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethanol_water,
        bias_cvs=[biased_distance, biased_torsion],
        model=None,  # type: ignore
    )

    with pytest.raises(
        ValueError,
        match="When using ADAPTIVE=GEOM, all CVs must have the same sigma value",
    ):
        meta_d_model.to_plumed(small_ethanol_water)



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import numpy as np
import plumedCommunications
from sys import stderr as log

# import plumedUtilities
log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

plumedInit = {
    "COMPONENTS": {
        "x": plumedCommunications.defaults.COMPONENT,
        "y": plumedCommunications.defaults.COMPONENT,
        "z": plumedCommunications.defaults.COMPONENT,
    }
}


def pydistInPair(action: plumedCommunications.PythonCVInterface):
    # NB: This is not a realistic case of using the neigbour list!!!
    # cvPY: PYCVINTERFACE GROUPA=1,4 IMPORT=pydistancegetAtPos CALCULATE=pydist

    atoms = action.getPositions()
    nl = action.getNeighbourList()
    assert nl.size == 3
    assert len(nl) == nl.size

    x, y, z = [atoms[pair[0]] - atoms[pair[1]] for pair in nl.getClosePairs()]

    zero = np.zeros(atoms.shape)
    # MUST work with tuple or lists
    return {"x": [x[0], zero], "y": (y[1], zero), "z": (z[2], zero)}



import unittest
import numpy as np

from utilities_for_test import *

import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class TestPyCV(unittest.TestCase):
    def test_nat(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("traj.xyz")
            )
            plmd = preparePlumed(num_atoms)
            cvPy = create_plumed_var(plmd, "cvPy", "PYCVINTERFACE IMPORT=atoms_number")
            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file
            with open("logfile", "w+") as of:
                # Now analyze the trajectory
                for step in range(0, num_frames):
                    of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n")
                    plmd.cmd("setStep", step)
                    plmd.cmd("setBox", box)
                    plmd.cmd("setMasses", masses)
                    plmd.cmd("setCharges", charges)
                    plmd.cmd("setPositions", traj[step])
                    plmd.cmd("setForces", forces)
                    plmd.cmd("setVirial", virial)
                    plmd.cmd("calc")

                    self.assertEqual(cvPy, 2)

    def test_atomPositions(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("traj.xyz")
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=1,4 IMPORT=pydistancegetAtPos CALCULATE=pydist",
            )
            cvCPP = create_plumed_var(plmd, "cvCPP", "DISTANCE ATOMS=1,4 NOPBC")

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file
            with open("logfile", "w+") as of:
                # Now analyze the trajectory
                for step in range(0, num_frames):
                    of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n")
                    plmd.cmd("setStep", step)
                    plmd.cmd("setBox", box)
                    plmd.cmd("setMasses", masses)
                    plmd.cmd("setCharges", charges)
                    plmd.cmd("setPositions", traj[step])
                    plmd.cmd("setForces", forces)
                    plmd.cmd("setVirial", virial)
                    plmd.cmd("calc")

                    np.testing.assert_almost_equal(cvPy, cvCPP, decimal=4)

    def test_atomPositionsPBC(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("traj.xyz")
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=1,4 IMPORT=pydistancePBCs CALCULATE=pydist",
            )
            cvCPP = create_plumed_var(plmd, "cvCPP", "DISTANCE ATOMS=1,4")

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file
            with open("logfile", "w+") as of:
                # Now analyze the trajectory
                for step in range(0, num_frames):
                    of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n")
                    plmd.cmd("setStep", step)
                    plmd.cmd("setBox", box)
                    plmd.cmd("setMasses", masses)
                    plmd.cmd("setCharges", charges)
                    plmd.cmd("setPositions", traj[step])
                    plmd.cmd("setForces", forces)
                    plmd.cmd("setVirial", virial)
                    plmd.cmd("calc")

                    np.testing.assert_almost_equal(cvPy, cvCPP, decimal=4)

    def test_newFrameNewAtomSTR(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("trajnewFrameNewAtom.xyz")
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=@mdatoms IMPORT=pycvPerFrameSTR CALCULATE=pydist PREPARE=changeAtom",
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file
            with open("logfile", "w+") as of:
                # Now analyze the trajectory
                for step in range(0, num_frames):
                    of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n")
                    plmd.cmd("setStep", step)
                    plmd.cmd("setBox", box)
                    plmd.cmd("setMasses", masses)
                    plmd.cmd("setCharges", charges)
                    plmd.cmd("setPositions", traj[step])
                    plmd.cmd("setForces", forces)
                    plmd.cmd("setVirial", virial)
                    plmd.cmd("calc")

                    np.testing.assert_almost_equal(cvPy, step + 1.0, decimal=4)

    def test_newFrameNewAtom(self):
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("trajnewFrameNewAtom.xyz")
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=@mdatoms IMPORT=pycvPerFrame CALCULATE=pydist PREPARE=changeAtom",
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")
            # Open an output file
            with open("logfile", "w+") as of:
                # Now analyze the trajectory
                for step in range(0, num_frames):
                    of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n")
                    plmd.cmd("setStep", step)
                    plmd.cmd("setBox", box)
                    plmd.cmd("setMasses", masses)
                    plmd.cmd("setCharges", charges)
                    plmd.cmd("setPositions", traj[step])
                    plmd.cmd("setForces", forces)
                    plmd.cmd("setVirial", virial)
                    plmd.cmd("calc")

                    np.testing.assert_almost_equal(cvPy, step + 1.0, decimal=4)

    def test_loadAmodule_and_persistData(self):
        """This test loads a module that is a directory and stores some data within plumed"""
        with cd(THIS_DIR):
            os.environ["PLUMED_MAXBACKUP"] = "0"
            traj, num_frames, num_atoms, box, virial, masses, forces, charges = (
                setUpTraj("trajnewFrameNewAtom.xyz")
            )
            plmd = preparePlumed(num_atoms)

            cvPy = create_plumed_var(
                plmd,
                "cvPy",
                "PYCVINTERFACE ATOMS=@mdatoms IMPORT=pycvPersistentData CALCULATE=pydist INIT=pyinit",
            )

            plmd.cmd("readInputLine", "PRINT FILE=colvar.out ARG=*")

            # Now analyze the trajectory
            for step in range(0, num_frames):
                plmd.cmd("setStep", step)
                plmd.cmd("setBox", box)
                plmd.cmd("setMasses", masses)
                plmd.cmd("setCharges", charges)
                plmd.cmd("setPositions", traj[step])
                plmd.cmd("setForces", forces)
                plmd.cmd("setVirial", virial)
                plmd.cmd("calc")
                # this cv sums the number of the step till now:
                # to future me: "//" is integer (floor) division
                np.testing.assert_almost_equal(
                    cvPy, ((step) * (step + 1)) // 2, decimal=4
                )


if __name__ == "__main__":
    # Output to four decimal places only
    np.set_printoptions(precision=4)
    unittest.main()



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import plumedCommunications
import pydoc

def plumedInit(_):
    #rather than the other doc example this creates a very ugly html manua;
    pydoc.writedoc(plumedCommunications)
    pydoc.writedoc(plumedCommunications.defaults)
    pydoc.writedoc(plumedCommunications.PythonCVInterface)
    return {"Value":plumedCommunications.defaults.COMPONENT_NODEV, "ATOMS":"1"}

def plumedCalculate(_):
    return 0



"""Tests for DistanceCV using the new API with VirtualAtom and flatten/pairwise parameters."""

import ase

import hillclimber as pn


def test_distance_cv_first_strategy(small_ethanol_water):
    """Test first strategy - distance between first ethanol and first water COMs."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),  # First ethanol
        x2=pn.VirtualAtom(x2_selector[0], "com"),  # First water
        prefix="d12",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    expected = [
        "d12_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_x2: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_x1,d12_x2",
    ]
    assert plumed_str == expected
    assert labels == ["d12"]


def test_distance_cv_all_pairs(small_ethanol_water):
    """Test all pairs strategy - all ethanol-water COM combinations."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector, "com"),  # All ethanols
        x2=pn.VirtualAtom(x2_selector, "com"),  # All waters
        prefix="d",
        pairwise="all",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    # Should create 2x2=4 distances
    expected = [
        "d_x1_0: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x1_1: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d_x2_0: COM ATOMS=19,20,21",
        "d_x2_1: COM ATOMS=22,23,24",
        "d_0: DISTANCE ATOMS=d_x1_0,d_x2_0",
        "d_1: DISTANCE ATOMS=d_x1_0,d_x2_1",
        "d_2: DISTANCE ATOMS=d_x1_1,d_x2_0",
        "d_3: DISTANCE ATOMS=d_x1_1,d_x2_1",
    ]
    assert plumed_str == expected
    assert labels == ["d_0", "d_1", "d_2", "d_3"]


def test_distance_cv_corresponding(small_ethanol_water):
    """Test diagonal/corresponding strategy - pair by index."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector, "com"),
        x2=pn.VirtualAtom(x2_selector, "com"),
        prefix="d",
        pairwise="diagonal",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    # Should create 2 distances (min of 2 ethanols and 2 waters)
    expected = [
        "d_x1_0: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x1_1: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d_x2_0: COM ATOMS=19,20,21",
        "d_x2_1: COM ATOMS=22,23,24",
        "d_0: DISTANCE ATOMS=d_x1_0,d_x2_0",
        "d_1: DISTANCE ATOMS=d_x1_1,d_x2_1",
    ]
    assert plumed_str == expected
    assert labels == ["d_0", "d_1"]


def test_distance_cv_first_to_all(small_ethanol_water):
    """Test first-to-all strategy - first ethanol to all waters."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "com"),  # First ethanol only
        x2=pn.VirtualAtom(x2_selector, "com"),  # All waters
        prefix="d",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    # Should create 2 distances (1 ethanol to 2 waters)
    expected = [
        "d_x1: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2_0: COM ATOMS=19,20,21",
        "d_x2_1: COM ATOMS=22,23,24",
        "d_0: DISTANCE ATOMS=d_x1,d_x2_0",
        "d_1: DISTANCE ATOMS=d_x1,d_x2_1",
    ]
    assert plumed_str == expected
    assert labels == ["d_0", "d_1"]


def test_distance_cv_single_atoms():
    """Test with single atom selections using flatten."""
    atoms = ase.Atoms("H2", positions=[[0, 0, 0], [1, 0, 0]])

    x1_selector = pn.IndexSelector(indices=[[0]])
    x2_selector = pn.IndexSelector(indices=[[1]])

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="h_h",
        flatten=True,
    )

    labels, plumed_str = distance_cv.to_plumed(atoms)

    expected = ["h_h: DISTANCE ATOMS=1,2"]
    assert plumed_str == expected
    assert labels == ["h_h"]


def test_distance_cv_first_atom_reduction(small_ethanol_water):
    """Test first atom selection using selector indexing."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    # Use selector indexing to get first atom of first group
    distance_cv = pn.DistanceCV(
        x1=x1_selector[0][0],  # First atom of first ethanol
        x2=x2_selector[0][0],  # First atom of first water
        prefix="d",
        flatten=True,
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    # Should use first atoms directly
    expected = ["d: DISTANCE ATOMS=1,19"]
    assert plumed_str == expected
    assert labels == ["d"]


def test_distance_cv_cog_reduction(small_ethanol_water):
    """Test center of geometry (COG) reduction."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=pn.VirtualAtom(x1_selector[0], "cog"),  # COG of first ethanol
        x2=pn.VirtualAtom(x2_selector[0], "cog"),  # COG of first water
        prefix="d",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    expected = [
        "d_x1: CENTER ATOMS=1,2,3,4,5,6,7,8,9",
        "d_x2: CENTER ATOMS=19,20,21",
        "d: DISTANCE ATOMS=d_x1,d_x2",
    ]
    assert plumed_str == expected
    assert labels == ["d"]


def test_distance_cv_no_virtual_sites(small_ethanol_water):
    """Test with flatten=True (direct atom lists, no virtual sites)."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector[0],
        x2=x2_selector[0],
        prefix="d",
        flatten=True,  # Use atoms directly
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethanol_water)

    # Should use flattened atom lists
    expected = ["d: DISTANCE ATOMS=1,2,3,4,5,6,7,8,9,19,20,21"]
    assert plumed_str == expected
    assert labels == ["d"]



import plumedCommunications as PLMD
import numpy

log = open("pydist.log", "w")

print("Imported my pydist+.", file=log)

initForF = {"Value": PLMD.defaults.COMPONENT_NODEV}


def function(action: PLMD.PythonFunction):
    arg = action.arguments()

    return numpy.abs(numpy.sqrt(arg[0] ** 2 + arg[1] ** 2 + arg[2] ** 2) - arg[3])



# This is only a partial example. This function does not compute the
# gradient, so it is useless for biasing. See the other regression
# tests for how to auto-grad.

# And, of course, one should not call slow functions (such as print)
# in the CV calculation.

import numpy as np
import plumedCommunications
from plumedCommunications.defaults import COMPONENT_NODEV
from sys import stderr as log
#log = open("pycv.log", "w")

print("Imported pycvPersistentData", file=log)

def pyinit(plmdAction: plumedCommunications.PythonCVInterface):
    print("Calling pyinit", file=log)
    plmdAction.log("---Calling pyinit---")
    plmdAction.lognl("Logging from Python :)")
    print(f"{plmdAction.data=}", file=log)
    plmdAction.data["pycv"]=0
    print(f"{plmdAction.data=}", file=log)
    return {"Value":COMPONENT_NODEV}

def pydist(plmdAction: plumedCommunications.PythonCVInterface):
    plmdAction.log("Calling pydist: ")
    plmdAction.lognl(plmdAction.getStep())
    print("Calling pydist", file=log)
    print(f"{plmdAction.data=}, {plmdAction.getStep()=}", file=log)
    
    plmdAction.data["pycv"]+=plmdAction.getStep()
    d=plmdAction.data["pycv"]
    return d



# The numbers calculated by the following python script are compared with those output from this PLUMED input
#
# MOLINFO STRUCTURE=template.pdb
# t1: TORSION ATOMS=@phi-2
# t2: TORSION ATOMS=@psi-2
# t3: TORSION ATOMS=@phi-3
# t4: TORSION ATOMS=@psi-3
# t5: TORSION ATOMS=@phi-4
# t6: TORSION ATOMS=@psi-4
# t7: TORSION ATOMS=@phi-5
# t8: TORSION ATOMS=@psi-5
# t9: TORSION ATOMS=@phi-6
# t10: TORSION ATOMS=@psi-6
# t11: TORSION ATOMS=@phi-7
# t12: TORSION ATOMS=@psi-7
# t13: TORSION ATOMS=@phi-8
# t14: TORSION ATOMS=@psi-8
# t15: TORSION ATOMS=@phi-9
# t16: TORSION ATOMS=@psi-9
# t17: TORSION ATOMS=@phi-10
# t18: TORSION ATOMS=@psi-10
# t19: TORSION ATOMS=@phi-11
# t20: TORSION ATOMS=@psi-11
# t21: TORSION ATOMS=@phi-12
# t22: TORSION ATOMS=@psi-12
# t23: TORSION ATOMS=@phi-13
# t24: TORSION ATOMS=@psi-13
# t25: TORSION ATOMS=@phi-14
# t26: TORSION ATOMS=@psi-14
# t27: TORSION ATOMS=@phi-15
# t28: TORSION ATOMS=@psi-15
# t29: TORSION ATOMS=@phi-16
# t30: TORSION ATOMS=@psi-16
# t31: TORSION ATOMS=@phi-17
# t32: TORSION ATOMS=@psi-17
#    
# PRINT ARG=t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32 FILE=colvar.ref FMT=%8.4f

import unittest
import numpy as np
import plumed
import os
from contextlib import contextmanager

@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)

def read_xyz(filename):
   xyz = open(filename)
   n_atoms = int(xyz.readline())
   title, trajectory = xyz.readline(), []
   while True :
      atom_type, coordinates = np.zeros(n_atoms).astype(str), np.zeros([n_atoms,3]) 
      for i in range(0,n_atoms) :
          line = xyz.readline()
          atom,x,y,z = line.split()
          atom_type[i]=atom
          coordinates[i,:]=np.array([x,y,z],dtype=np.float64)
      trajectory.append( coordinates )
      nextline = xyz.readline()
      if( nextline=="" ) : break
      c_atoms = int(nextline)
      if( c_atoms!=n_atoms ) : break 
      title = xyz.readline()
   xyz.close()
   return trajectory

def create_plumed_var( p, name, command ):
   p.cmd("readInputLine", name + ": " + command )
   shape = np.zeros( 1, dtype=np.int_ )
   p.cmd("getDataRank " + name, shape )
   data = np.zeros((1))
   p.cmd("setMemoryForData " + name, data )
   return data

class Test(unittest.TestCase):
  def runtest(self):
    os.system('rm -f bck.*')
    # Output to four decimal places only
    np.set_printoptions(precision=4)
    # Read trajectory
    traj = read_xyz("traj.xyz")
    num_frames = len(traj)
    num_atoms = traj[0].shape[0]
    
    # Create arrays for stuff
    box=np.diag(12.41642*np.ones(3,dtype=np.float64))
    virial=np.zeros((3,3),dtype=np.float64)
    masses=np.ones(num_atoms,dtype=np.float64)
    forces=np.random.rand(num_atoms,3)
    charges=np.zeros(num_atoms,dtype=np.float64)
    
    # Create PLUMED object and read input
    p = plumed.Plumed()

    # not really needed, used to check https://github.com/plumed/plumed2/issues/916
    plumed_version = np.zeros(1, dtype=np.intc)
    p.cmd( "getApiVersion", plumed_version)

    p.cmd("setMDEngine","python")
    p.cmd("setTimestep", 1.)
    p.cmd("setKbT", 1.)
    p.cmd("setNatoms",num_atoms)
    p.cmd("setLogFile","test.log")
    p.cmd("init")
    p.cmd("readInputLine","MOLINFO STRUCTURE=template.pdb")
    t1 = create_plumed_var( p, "t1", "TORSION ATOMS=@phi-2" ) 
    t2 = create_plumed_var( p, "t2", "TORSION ATOMS=@psi-2" )
    t3 = create_plumed_var( p, "t3", "TORSION ATOMS=@phi-3" )
    t4 = create_plumed_var( p, "t4", "TORSION ATOMS=@psi-3" )
    t5 = create_plumed_var( p, "t5", "TORSION ATOMS=@phi-4" )
    t6 = create_plumed_var( p, "t6", "TORSION ATOMS=@psi-4" )
    t7 = create_plumed_var( p, "t7", "TORSION ATOMS=@phi-5" )
    t8 = create_plumed_var( p, "t8", "TORSION ATOMS=@psi-5" )
    t9 = create_plumed_var( p, "t9", "TORSION ATOMS=@phi-6" )
    t10 = create_plumed_var( p, "t10", "TORSION ATOMS=@psi-6" )
    t11 = create_plumed_var( p, "t11", "TORSION ATOMS=@phi-7" )
    t12 = create_plumed_var( p, "t12", "TORSION ATOMS=@psi-7" )
    t13 = create_plumed_var( p, "t13", "TORSION ATOMS=@phi-8" )
    t14 = create_plumed_var( p, "t14", "TORSION ATOMS=@psi-8" )
    t15 = create_plumed_var( p, "t15", "TORSION ATOMS=@phi-9" )
    t16 = create_plumed_var( p, "t16", "TORSION ATOMS=@psi-9" )
    t17 = create_plumed_var( p, "t17", "TORSION ATOMS=@phi-10" )
    t18 = create_plumed_var( p, "t18", "TORSION ATOMS=@psi-10" )
    t19 = create_plumed_var( p, "t19", "TORSION ATOMS=@phi-11" )
    t20 = create_plumed_var( p, "t20", "TORSION ATOMS=@psi-11" )
    t21 = create_plumed_var( p, "t21", "TORSION ATOMS=@phi-12" )
    t22 = create_plumed_var( p, "t22", "TORSION ATOMS=@psi-12" )
    t23 = create_plumed_var( p, "t23", "TORSION ATOMS=@phi-13" )
    t24 = create_plumed_var( p, "t24", "TORSION ATOMS=@psi-13" )
    t25 = create_plumed_var( p, "t25", "TORSION ATOMS=@phi-14" )
    t26 = create_plumed_var( p, "t26", "TORSION ATOMS=@psi-14" )
    t27 = create_plumed_var( p, "t27", "TORSION ATOMS=@phi-15" )
    t28 = create_plumed_var( p, "t28", "TORSION ATOMS=@psi-15" )
    t29 = create_plumed_var( p, "t29", "TORSION ATOMS=@phi-16" )
    t30 = create_plumed_var( p, "t30", "TORSION ATOMS=@psi-16" )
    t31 = create_plumed_var( p, "t31", "TORSION ATOMS=@phi-17" )
    t32 = create_plumed_var( p, "t32", "TORSION ATOMS=@psi-17" )
     
    # Read in the correct answers that were calculated directly using PLUMED
    correct_torsions = np.loadtxt("colvar.ref")
    # Open an output file
    of = open("logfile", "w+")
    
    # Now analyze the trajectory
    for step in range(0,num_frames) :
        of.write("RUNNING ANALYSIS FOR STEP " + str(step) + "\n" )
        p.cmd("setStep",step )
        p.cmd("setBox",box )
        p.cmd("setMasses", masses )
        p.cmd("setCharges", charges )
        p.cmd("setPositions", traj[step])
        p.cmd("setForces", forces )
        p.cmd("setVirial", virial )
        p.cmd("calc")
        bias = np.zeros((1),dtype=np.float64)
        p.cmd("getBias", bias )
        variables = np.array([t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32]).ravel()
        zeros = variables - correct_torsions[step,1:]
        for data in zeros :
            self.assertAlmostEqual(data,0.0,places=4)
    of.close()

  def test(self):
    with cd('test/'):
        self.runtest()

if __name__ == "__main__":
    unittest.main()

